import torch
import torch.nn as nn
import torchvision.models as models
import torchvision.transforms as transforms
from torch.autograd import Variable
from PIL import Image
import cv2 as cv

# alt options are alexnet, vgg11, and densenet
model = models.resnet18(pretrained=True)
layer = model._modules.get('avgpool')
model.eval()

scaler = transforms.Resize((224, 224))
# expected normalization for pretrained pytorch networks
normalize = transforms.Normalize(
    mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
to_tensor = transforms.ToTensor()


def generate_feature_vector(image):
    image = Image.fromarray(image)
    transformed = Variable(normalize(to_tensor(scaler(image))).unsqueeze(0))
    embedding = torch.zeros(512)  # resnet18 uses 512 features

    def copy_layer_output(module, input, output):
        embedding.copy_(output.data.reshape(output.data.size(1)))

    # on the forward pass in the referenced layer, run the copy layer output function
    handle = layer.register_forward_hook(copy_layer_output)

    model(transformed)

    handle.remove()

    return embedding


def group(vectors, threshold=0.9):
    cos = nn.CosineSimilarity(dim=1, eps=1e-6)
    scores = []

    for i, u in enumerate(vectors):
        if i != len(vectors) - 1:
            scores.append([])

        for j, v in enumerate(vectors):
            if j > i:
                scores[i].append(cos(u.unsqueeze(0), v.unsqueeze(0)))
    print(scores)
    groups = []
    for i, row in enumerate(scores):
        for j, element in enumerate(row):
            j = j + i + 1
            if element >= threshold:
                found = False
                for k, group in enumerate(groups):
                    if i in group or j in group:
                        groups[k].add(i)
                        groups[k].add(j)
                        found = True
                if not found:
                    # print({i, j})
                    groups.append({i, j})
            else:
                found_i = False
                for k, group in enumerate(groups):
                    if i in group:
                        found_i = True
                if not found_i:
                    groups.append({i})
    found_last = False
    for k, group in enumerate(groups):
        if len(vectors) - 1 in group:
            found_last = True
    if not found_last:
        groups.append({len(vectors) - 1})
    # print(scores)
    # print(groups)
    return groups


# u = generate_feature_vector(cv.imread('../resources/examples/IMG_1663.jpg'))
# v = generate_feature_vector(cv.imread('../resources/examples/IMG_1665.jpg'))
# w = generate_feature_vector(cv.imread('../resources/examples/IMG_1667.jpg'))
# x = generate_feature_vector(cv.imread('../resources/examples/IMG_1668.jpg'))
# y = generate_feature_vector(cv.imread('../resources/examples/IMG_1669.jpg'))


# group([u, v, w, x, y])



from os import read
import cv2 as cv
# import mtcnn
import mediapipe as mp
from matplotlib import pyplot as plt


FACE_DETECTOR = mp.solutions.face_detection.FaceDetection(
    min_detection_confidence=0.4)


def read_img(filename):
    return plt.imread(filename)


def save_img(filename, img):
    cv.imwrite(filename, img)


def save_img_colored(filename, img):
    cv.imwrite(filename, cv.cvtColor(img, cv.COLOR_RGB2BGR))


def show_img(image):
    cv.imshow('Image', image)
    cv.waitKey(0)


def detect_faces(image):
    if image is None:
        raise Exception('No image found.')

    return FACE_DETECTOR.process(image).detections


def get_subject_bounds(face_info):
    return face_info.location_data.relative_bounding_box


# def correct_eye_bounds(top_left, bottom_right):
#     new_top_left = min(top_left[0], bottom_right[0]), min(
#         top_left[1], bottom_right[1])
#     new_bottom_right = max(top_left[0], bottom_right[0]), max(
#         top_left[1], bottom_right[1])
#     return new_top_left, new_bottom_right


def is_left_eye_higher(face_info):
    return face_info['keypoints']['left_eye'][1] < face_info['keypoints']['right_eye'][1]


def get_eye_bounds(face_info):
    # Average eye diameter is roughly 23 mm: Bekerman, Inessa & Gottlieb, Paul & Vaiman, Michael. (2014). Variations in Eyeball Diameters of the Healthy Adults. Journal of ophthalmology. 2014. 503645. 10.1155/2014/503645. \
    # Average face width is roughly 147.6 mm: DU LL, Wang LM, Zhuang Z. [Measurement and analysis of human head-face dimensions]. Zhonghua Lao Dong Wei Sheng Zhi Ye Bing Za Zhi. 2008 May;26(5):266-70. Chinese. PMID: 18727867.

    # Thus, the width of each eye should be roughly 16% of the face width
    # This function computes eye bounds by extending 16% of the face width upward and leftward from the left eye and the same downward and rightward from the right eye
    # The extension is used as 16% rather than 8% in either direction to allow cases where the innermost edge of the eye is detected as the eye
    extension = 0.16

    face_bounds = get_subject_bounds(face_info)
    _, _, face_width, _ = face_bounds

    left_higher = is_left_eye_higher(face_info)

    if left_higher:
        top_left = int(face_info['keypoints']['left_eye'][0] - (
            face_width * extension)), int(face_info['keypoints']['left_eye'][1] - (face_width * extension))
        bottom_right = int(face_info['keypoints']['right_eye'][0] + (
            face_width * extension)), int(face_info['keypoints']['right_eye'][1] + (face_width * extension))
    else:
        top_left = int(face_info['keypoints']['left_eye'][0] - (
            face_width * extension)), int(face_info['keypoints']['right_eye'][1] - (face_width * extension))
        bottom_right = int(face_info['keypoints']['right_eye'][0] + (
            face_width * extension)), int(face_info['keypoints']['left_eye'][1] + (face_width * extension))

    return (top_left, bottom_right)


def show_point(image, point):
    image = cv.circle(image, point, radius=10, color=(0, 0, 255), thickness=-1)
    show_img(image)


def show_rect(image, top_left, bottom_right):
    cv.rectangle(
        image, (top_left[0], top_left[1]), (bottom_right[0], bottom_right[1]), (0, 255, 0), 1)
    show_img(image)


def crop_subjects(image):
    faces_info = detect_faces(image)

    im_height, im_width, _ = image.shape

    subjects = []

    if not faces_info:
        return []

    for face in faces_info:
        print(face)
        bounds = get_subject_bounds(face)

        x, y, w, h = int(bounds.xmin * im_width), int(bounds.ymin * im_height), int(
            bounds.width * im_width), int(bounds.height * im_height)
        sub = image[y:y + h, x:x + w]
        sub = cv.cvtColor(sub, cv.COLOR_BGR2RGB)
        subjects.append(sub)
    return subjects


# def show_bounding_boxes(image, bounding_boxes):
#     for box in bounding_boxes:
#         x, y, w, h = box['box']
#         cv.rectangle(image, (x, y), (x + w, y + h), (0, 255, 0), 5)
#     show_img(image)

def get_eye_frame_from_img(filename):
    img = read_img(filename)
    faces_info = detect_faces(img)
    for face in faces_info:
        yield get_eye_bounds(face)


def main():
    img = read_img('./bestOf/resources/examples/fourfaces.jpg')
    subs = crop_subjects(img)
    for sub in subs:
        show_img(sub)


if __name__ == '__main__':
    main()



import os
import identifyPeople


def main():
    blinkset_path = '../resources/blinkset/'
    eyeset_path = '../resources/eyeset_colored/'
    for folder_name in os.listdir(blinkset_path):

        # use the following code block when only updating one class, comment out pass or continue as apt
        if folder_name == 'blinkers':
            continue
            # pass
        else:
            # continue
            pass

        for img_name in os.listdir(blinkset_path + folder_name + '/'):
            try:
                eyes_gen = identifyPeople.get_eye_frame_from_img(
                    blinkset_path + folder_name + '/' + img_name)
                eyes = None
                for eye_box in eyes_gen:
                    eyes = eye_box
                    break
                if not eyes:
                    continue
                image = identifyPeople.read_img(
                    blinkset_path + folder_name + '/' + img_name)
                image = image[eyes[0][1]:eyes[1][1], eyes[0][0]:eyes[1][0]]
                identifyPeople.save_img_colored(
                    eyeset_path + folder_name + '/' + img_name, image)
            except Exception as e:
                print(e)
                continue


if __name__ == '__main__':
    main()



import torch
from torch.utils.data import Dataset

# Custom dataset model for the data
class custom_dataset(Dataset):
    def __init__(self, all_x, all_y, transform, target_transform=None):
        self.all_x = all_x
        self.all_y = all_y
        self.transform = transform
        self.target_transform = target_transform
        
    def __len__(self):
        return len(self.all_x)
    
    def __getitem__(self, index):
        image = self.all_x[index]
        label = self.all_y[index]
        image = self.transform(image)
        
        if self.target_transform:
            label = self.target_transform(label)

        return image, label



import cv2 as cv
from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
import sys
import os
print("Starting...")

# sys.path.insert(1, 'bestOf/backend')


import bestOf.backend.similarity as similarity
from bestOf.backend.classDefinitions import BlinkAndCropNet, cropped_model
import bestOf.backend.blinkDetector as blinkDetector
import bestOf.backend.cropDetector as cropDetector
import bestOf.backend.identifyPeople as identifyPeople


def loadImages(filenames):
    for filename in filenames:
        yield identifyPeople.read_img(filename)


IMAGELIST = []
GROUPS = []

# Reference Source 1: https://learndataanalysis.org/how-to-pass-data-from-one-window-to-another-pyqt5-tutorial/


def main():
    class processingResults(QWidget):
        def __init__(self):
            super().__init__()
            self.options()
            self.setWindowTitle("Best Of")
            self.setGeometry(1550, 800, 500, 500)
            self.layout = QVBoxLayout()
            self.layout.addWidget(self.des)
            self.setLayout(self.layout)

        def display(self):
            self.__init__()
            self.show()

        def options(self):
            self.des = QGroupBox("Images Selected.",
                                 alignment=QtCore.Qt.AlignCenter)
            vert = QVBoxLayout()
            layout = QHBoxLayout()
            self.redo = QPushButton('&Run Again')
            self.redo.clicked.connect(self.imageProcessing)
            layout.addWidget(self.redo)
            self.back = QPushButton('&Back to Main Menu')
            self.back.clicked.connect(self.close)
            layout.addWidget(self.back)
            info = IMAGELIST
            groups = GROUPS
            if len(info) == 0:
                return
            print(info)
            print(groups)
            for group in groups:
                h = QHBoxLayout()
                for index in group:
                    lab = QLabel()
                    for el in info:
                        if el[0] == index:
                            lab.setPixmap(QPixmap(el[1]))
                    h.addWidget(lab)
                vert.addLayout(h)
            vert.addLayout(layout)
            self.des.setLayout(vert)

        def imageProcessing(self):
            # Code from backend goes here...
            print("Redoing image selection.")

    class settingsMenu(QWidget):
        def __init__(self):
            super().__init__()
            self.options()
            self.setWindowTitle("Settings Menu")
            self.setGeometry(1550, 800, 500, 500)
            self.layout = QVBoxLayout()
            self.layout.addWidget(self.des)
            self.setLayout(self.layout)

        def display(self):
            self.show()

        def options(self):
            self.des = QGroupBox("Apply the following criteria for image refinement.",
                                 alignment=QtCore.Qt.AlignCenter)
            layout = QHBoxLayout()

            self.closeButton = QPushButton('&Close Settings')
            self.closeButton.clicked.connect(self.close)
            layout.addWidget(self.closeButton)
            self.des.setLayout(layout)

    class mainWindow(QWidget):
        def __init__(self):
            super().__init__()
            self.setWindowTitle("Best Of: Main Menu")
            self.resize(800, 800)
            self.setGeometry(1550, 800, 500, 500)
            self.nextWindow = settingsMenu()
            self.process = processingResults()
            self.makeUI()
            self.layout = QVBoxLayout()
            self.layout.addWidget(self.firstText)
            self.setLayout(self.layout)

        def makeUI(self):
            self.firstText = QGroupBox("Welcome, browse your computer for images to start picture analysis.",
                                       alignment=QtCore.Qt.AlignCenter)
            layout = QHBoxLayout()
            self.addImage = QPushButton('&Add Images')
            self.addImage.clicked.connect(self.getFiles)
            layout.addWidget(self.addImage)
            self.processImage = QPushButton('&Process Image(s)')
            self.processImage.clicked.connect(self.process.display)
            layout.addWidget(self.processImage)
            self.settings = QPushButton('&More Settings')
            self.settings.clicked.connect(self.nextWindow.display)
            layout.addWidget(self.settings)
            self.firstText.setLayout(layout)

        # Grabs the files for image selection... work in progress...

        def getFiles(self):
            global IMAGELIST
            global GROUPS
            file = QFileDialog.getOpenFileNames(
                self, 'Add Files', QtCore.QDir.rootPath())
            files = ""
            layout = QVBoxLayout
            image_generator = loadImages(list(file[0]))
            vectors = []
            for image in image_generator:
                v = similarity.generate_feature_vector(image)
                vectors.append(v)

            threshold = 0.8  # this should be grabbed from whatever the user set it to in the settings
            groups = similarity.group(vectors, threshold=threshold)
            print(groups)
            GROUPS = groups

            image_generator = loadImages(list(file[0]))

            bestof_scores = []

            for image in image_generator:
                # print('Scanning Image...')
                subjects = identifyPeople.crop_subjects(image)
                print("len of subjects", len(subjects))
                blinks = 0
                crops = 0
                for sub in subjects:
                    # identifyPeople.show_img(sub)
                    # print('Scanning Subject...')
                    if blinkDetector.test(sub):
                        blinks += 1
                    if cropDetector.test(sub):
                        crops += 1

                if len(subjects) == 0:
                    bestof_scores.append(0)
                    continue

                blink_score = (len(subjects) - blinks) / len(subjects)
                crop_score = (len(subjects) - crops) / len(subjects)

                bestof_score = blink_score + crop_score
                bestof_scores.append(bestof_score)
            final = zip(range(len(file[0])), file[0], bestof_scores)

            final = sorted(final, key=lambda x: x[2], reverse=True)
            print(final)
            IMAGELIST = final

    newApp = QApplication(sys.argv)  # Creates application class
    wind = mainWindow()
    wind.show()
    newApp.exec()  # Executes app with the window


if __name__ == '__main__':
    main()



from setuptools import setup

setup(
    name="bestOf",
    version="0.0.51",
    description="A module that uses machine learning to chose the best image in groups",
    packages=["bestOf", "bestOf.backend",
              "bestOf.backend.saved_models", "bestOf.frontend", ],
    url="https://github.com/apangasa/bestof",
    install_requires=["torch",
                      "torchvision",
                      "mediapipe",
                      "PyQt5",
                      "Pillow",
                      "matplotlib",
                      "opencv-python"],
    include_package_data=True,

    entry_points={
        'console_scripts': ['bestOf = bestOf.frontend.frontend:main'],
    }

)



import math
import cv2 as cv
import numpy as np
import os
import torch
import torchvision
from torchvision import transforms
import torch.optim as optim
import torch.nn as nn
import torch.nn.functional as fn
from torch.utils.data import DataLoader, Dataset, random_split
from bestOf.backend.createCroppedFaceDataset import custom_dataset


class BlinkAndCropNet(torch.nn.Module):
    def __init__(self):
        super(BlinkAndCropNet, self).__init__()
        self.conv1 = nn.Conv2d(3, 32, 3)
        self.conv2 = nn.Conv2d(32, 32, 3)
        self.pool1 = nn.MaxPool2d(2)

        self.drop1 = nn.Dropout2d(0.25)

        self.conv3 = nn.Conv2d(32, 64, 3)
        self.pool2 = nn.MaxPool2d(2)

        self.drop2 = nn.Dropout2d(0.25)

        self.dense1 = nn.Linear(20736, 128)
        self.dense2 = nn.Linear(128, 64)
        self.dense3 = nn.Linear(64, 1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        x = fn.relu(self.conv2(fn.relu(self.conv1(x))))
        x = self.drop1(self.pool1(x))

        x = fn.relu(self.conv3(x))
        x = self.drop2(self.pool2(x))
        x = torch.flatten(x, 1)

        x = fn.relu(self.dense1(x))
        x = fn.relu(self.dense2(x))
        x = self.dense3(x)
        x = self.sigmoid(x)
        return x


def load_dataset(img_height, img_width):
    print("Loading images into dataset...")

    # Loads all images from lfw and lfw_cut with opencv and stores it in all_x,
    # stores all labels in all_y
    nonblinkpath = "../resources/blinkset/nonblinkers"
    blinkpath = "../resources/blinkset/blinkers"

    all_x = []
    all_y = []
    img_count = 0
    # Load all images from lfw with label in all_x and all_y
    # for folder in os.listdir(nonblinkpath):
    #folder_path = os.path.join(nonblinkpath, folder)

    for img in os.listdir(nonblinkpath):
        image = cv.imread(os.path.join(nonblinkpath, img))

        if image.shape[0] < img_height or image.shape[1] < img_width:
            continue

        image = cv.cvtColor(image, cv.COLOR_BGR2RGB)
        image = cv.resize(image, (img_height, img_width),
                          interpolation=cv.INTER_AREA)
        img_count += 1
        all_x.append(image)
        all_y.append(0.0)

    # for folder in os.listdir(blinkpath):
        #folder_path = os.path.join(blinkpath, folder)

    for img in os.listdir(blinkpath):
        image = cv.imread(os.path.join(blinkpath, img))

        if image.shape[0] < img_height or image.shape[1] < img_width:
            continue

        image = cv.cvtColor(image, cv.COLOR_BGR2RGB)
        image = cv.resize(image, (img_height, img_width),
                          interpolation=cv.INTER_AREA)
        img_count += 1
        all_x.append(image)
        all_y.append(1.0)

    # Convert to np array to change to tensor
    all_x = np.asarray(all_x)
    all_y = np.asarray(all_y)
    print("Finished loading all images! Loaded " + str(img_count) + " images!\n")
    return all_x, all_y


def train(all_x, all_y, split, device):
    # Define basic variables and parameters
    learning_rate = 0.00005
    batch_size = 64
    epoch = 11
    model_path = "./saved_models/blinkingFaces.pth"

    # Create all of the dataloaders and models, then train it with train_loader
    transform = transforms.Compose(
        [transforms.ToTensor()])

    full_dataset = custom_dataset(all_x, all_y, transform)
    train_len = math.floor(len(all_x) * split)
    val_len = len(all_x) - train_len
    train_dataset, val_dataset = random_split(
        full_dataset, [train_len, val_len])

    print(str(train_dataset.__len__))

    train_loader = DataLoader(
        train_dataset, batch_size=batch_size, shuffle=True)
    val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False)

    model = BlinkAndCropNet().to(device)
    criteria = nn.BCELoss()
    opt = optim.Adam(model.parameters(), learning_rate)

    n_total_steps = len(train_loader)

    if torch.cuda.is_available():
        print("Starting the training using the GPU")
    else:
        print("Starting the training using the CPU, this'll take a while")

    for epoch_num in range(epoch):
        for i, (images, labels) in enumerate(train_loader):
            images = images.to(device)
            labels = labels.to(device)

            outputs = model(images)
            loss = criteria(outputs[:, 0], labels.float())

            opt.zero_grad()
            loss.backward()
            opt.step()

            if (i + 1) % 20 == 0:
                print(
                    f"Epoch: [{epoch_num+1}/{epoch}], Step: [{i+1}/{n_total_steps}], Loss: {loss.item():.4f}")
    print("Successfully finished training!\n")
    torch.save(model.state_dict(), model_path)

    return model, val_loader


def validation(model, val_loader, device):
    # Validate the model
    print("Validating model...")

    model.eval()
    with torch.no_grad():
        correct = 0
        samples = 0
        acc = 0.0

        with torch.no_grad():
            for images, labels in val_loader:
                images = images.to(device=device)
                labels = labels.to(device=device)

                scores = model(images)
                scores = torch.round(scores)
                scores = scores[:, 0]

                correct += (scores == labels).sum()
                samples += scores.size(0)

            acc = (correct / samples) * 100
            print("Accuracy: " + str(acc) + "\n")


if __name__ == "__main__":
    img_height = 80
    img_width = 80
    train_val_split = 0.75
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    all_x, all_y = load_dataset(img_height, img_width)
    model, val_loader = train(all_x, all_y, train_val_split, device)
    validation(model, val_loader, device)
    print("Done with creating the face cropped detection model!")
